#include <asm/param.h>
#include <asm/types.h>
#include <linux/pkt_sched.h>
#include <linux/pkt_cls.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>


/* Commands */
#define LQ_add		1
#define LQ_del		2
#define LQ_change	3


/* Object types */
#define	LQ_qdisc		1
#define	LQ_class		2
#define	LQ_filter		3


/* Qdisc type */
#define LQ_q_null		0
#define	LQ_q_htb		LQ_q_null + 1
#define	LQ_q_sfq		LQ_q_null + 2
#define	LQ_q_pfifo		LQ_q_null + 3
#define	LQ_q_bfifo		LQ_q_null + 4
#define	LQ_q_prio		LQ_q_null + 5
#define	LQ_q_dup		LQ_q_null + 6
#define	LQ_q_ooo		LQ_q_null + 7
#define	LQ_q_esfq		LQ_q_null + 8

/* Class type */
#define LQ_c_null		0
#define LQ_c_htb		LQ_c_null + 1
#define LQ_c_tbf		LQ_c_null + 2

/* Filter type */
#define LQ_f_null		0
#define LQ_f_u32		LQ_f_null + 1
#define LQ_f_fw			LQ_f_null + 2
#define LQ_f_route		LQ_f_null + 3
#define LQ_f_tcindex		LQ_f_null + 4

/* Operations */
#define LQ_op_add		0
#define LQ_op_del		LQ_op_add + 1
#define LQ_op_change		LQ_op_add + 2
#define LQ_op_replace		LQ_op_add + 3


typedef struct {
	int ifindex;
	u_int32_t handle;
	u_int32_t parent;
	u_int64_t bytes;
	u_int32_t packets;
	u_int32_t drops;
	u_int32_t overlimits;
	u_int32_t bps;
	u_int32_t pps;
	u_int32_t qlen;
	u_int32_t backlog;

	u_int32_t rate;
	u_int32_t ceil;
} lq_stats;


/* Main struct */
typedef struct {
	unsigned int	magic;

	short		type;		/* qdisc, filter or class */
	short		subtype;	/* what kind of qdisc, filter or class */

	__u32		parent;
	__u32		id;

	int		ifindex;	/* interface index */

	struct tc_stats	stats;		/* statistics */

	/* specific pointer */
	union {
		/* qdisc */
		union {
			/* htb */
			struct {
				struct tc_htb_glob	glob;
				struct tc_htb_opt	opt;
				struct tc_htb_xstats	xstats;
			} htb;

			/* sfq */
			struct {
				struct tc_sfq_qopt	opt;
			} sfq;

			/* fifo */
			struct {
				struct tc_fifo_qopt	opt;
			} fifo;

			struct {
				struct tc_prio_qopt	opt;
			} prio;
		} q;

		/* class */
		union {
			/* htb */
			struct {
				struct tc_htb_opt	opt;
				struct tc_htb_xstats	xstats;
			} htb;
		} c;


		/* filter */
		struct {
			__u16		prio;
			__u16		protocol;
			__u32		classid;

			union {
				/* fw */
				/* nothing */

				/* u32 */
				struct {
					unsigned		divisor;
					unsigned		link;
					unsigned		ht;
					__u32			order;

					__u32			mask, at;

					struct tc_u32_sel	sel;
					struct tc_u32_key	keys[16];
				} u32;
			} f2;
		} f;
	} data;
} lq;

/* Functions */
extern char		*lq_subtype(lq *O);
extern char		*lq_type(lq *O);
extern char		*lq_dump_stats(lq *O);
extern int		lq_read_psched(void);
extern int		lq_compute_rtable(unsigned long bps, __u32 *rtab, int cell_log, unsigned int mtu, unsigned int mpu);
extern unsigned long	lq_xmittime(unsigned long rate, unsigned long size);
extern void		lq_print_rate (char *buf, unsigned int len, unsigned long rate);
extern unsigned long	lq_tick2usec(long tick);
extern void		lq_rate2string(u_int64_t rate, char *str, int size);
